/* Soot - a J*va Optimization Framework
 * Copyright (C) 1997-1999 Raja Vallee-Rai
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the Sable Research Group and others 1997-1999.  
 * See the 'credits' file distributed with Soot for the complete list of
 * contributors.  (Soot is distributed at http://www.sable.mcgill.ca/soot)
 */

/* Reference Version: $SootVersion: 1.0.0.dev.1 $ */




package soot;

import soot.*;
import soot.util.*;
import java.util.*;

/** Provides default implementations for the methods in Unit. */
public abstract class AbstractUnit extends AbstractHost implements Unit 
{
    static Map allMapToUnnamed = Collections.unmodifiableMap(new AllMapTo("<unnamed>"));

    /** Returns a deep clone of this object. */
    public abstract Object clone();
    
    /** Returns a list of Boxes containing Values used in this Unit.
     * The list of boxes is dynamically updated as the structure changes.
     * Note that they are returned in usual evaluation order.
     * (this is important for aggregation)
     */
    public List getUseBoxes()
    {
        return emptyList;
    }

    /** Returns a list of Boxes containing Values defined in this Unit.
     * The list of boxes is dynamically updated as the structure changes.
     */
    public List getDefBoxes()
    {
        return emptyList;
    }


    /** Returns a list of Boxes containing Units defined in this Unit; typically
     * branch targets.
     * The list of boxes is dynamically updated as the structure changes.
     */
    public List getUnitBoxes()
    {
        return emptyList;
    }

    /** Canonical AbstractUnit.emptyList list. */
    static final public List emptyList = Collections.unmodifiableList(new ArrayList());

    /** List of UnitBoxes pointing to this Unit. */
    List boxesPointingToThis = new ArrayList();

    /** List of ValueBoxes contained in this Unit. */
    List valueBoxes = null;

    /** Returns a list of Boxes pointing to this Unit. */
    public List getBoxesPointingToThis()
    {
        return boxesPointingToThis;
    }

    /** Returns a list of ValueBoxes, either used or defined in this Unit. */
    public List getUseAndDefBoxes()
    {
        valueBoxes = new ArrayList();

        valueBoxes.addAll(getUseBoxes());
        valueBoxes.addAll(getDefBoxes());

        valueBoxes = Collections.unmodifiableList(valueBoxes);

        return valueBoxes;
    }

    /** Used to implement the Switchable construct. */
    public void apply(Switch sw)
    {
    }

    /** Returns a brief description of this object. */
    public String toBriefString()
    {
        return toString(true, allMapToUnnamed, "");
    }
    
    /** Returns a brief description of this object, using the given Map. */
    public String toBriefString(Map stmtToName)
    {
        return toString(true, stmtToName, "");
    }
    
    /** Returns a brief description of this object, using the given indentation. */
    public String toBriefString(String indentation)
    {
        return toString(true, allMapToUnnamed, indentation);
    }
    
    /** Returns a brief description of this object, using the given Map and indentation. */
    public String toBriefString(Map stmtToName, String indentation)
    {
        return toString(true, stmtToName, indentation);
    }
    
    /** Returns a textual representation of this object. */
    public String toString()
    {
        return toString(false, allMapToUnnamed, "");
    }
    
    /** Returns a textual representation of this object, using the given Map. */
    public String toString(Map stmtToName)
    {
        return toString(false, stmtToName, "");
    }
    
    /** Returns a textual representation of this object, using the given indentation. */
    public String toString(String indentation)
    {
        return toString(false, allMapToUnnamed, indentation);
    }
    
    /** Returns a textual representation of this object, using the given Map and indentation. */
    public String toString(Map stmtToName, String indentation)
    {
        return toString(false, stmtToName, indentation);
    }
    
    /** Returns a textual representation of this object, with the given briefness, Map and indentation. */
    abstract protected String toString(boolean isBrief, Map stmtToName, String indentation);

    static class AllMapTo extends AbstractMap
    {
        Object dest;
        
        public AllMapTo(Object dest)
        {
            this.dest = dest;
        }
        
        public Object get(Object key)
        {
            return dest;
        }
        
        public Set entrySet()
        {
            throw new UnsupportedOperationException();
        }
    }

    /** Adjusts UnitBoxes pointing to this Unit to a newLocation. */
    public void redirectJumpsToThisTo(Unit newLocation)
    {
        List boxesPointing = this.getBoxesPointingToThis();

        Object[] boxes = boxesPointing.toArray();
        // important to change this to an array to have a static copy
        
        for(int i = 0; i < boxes.length; i++)
        {
            UnitBox box = (UnitBox) boxes[i];

            if(box.getUnit() != this)
                throw new RuntimeException("Something weird's happening");

            box.setUnit(newLocation);
        }

    }


    
}

